@@ -9,6 +9,7 @@ import ast
 import traceback
 from typing import List, Optional, Union, Any, Dict, Tuple
 from .docker_manager import get_docker_manager
+from .subprocess_executor import get_subprocess_executor, cleanup_subprocess_executor
 from .languages import language_manager
 from .security import validate_code, sanitize_code
 from ..config.settings import settings
@@ -28,6 +29,7 @@ class CodeExecutor:
     def __init__(self):
         """Initialize the code executor."""
         self.docker_manager = get_docker_manager()
+        self.subprocess_executor = get_subprocess_executor()
     
     def execute(self,
                 inputs: Optional[List[Any]] = None,
@@ -67,12 +69,12 @@ class CodeExecutor:
             execution_mode = self._determine_execution_mode(inputs, outputs, code, answer_fn, mode)
 
             if execution_mode == "answer":
-                return self._execute_input_output_mode(inputs, outputs, code, language, timeout)
+                return self._execute_input_output_mode_subprocess(inputs, outputs, code, language, timeout)
             elif execution_mode == "function":
-                return self._execute_function_comparison_mode(answer_fn, code, language, timeout)
+                return self._execute_function_comparison_mode_subprocess(answer_fn, code, language, timeout)
             else:
                 # Simple execution mode
-                return self._execute_simple_mode(code, language, timeout)
+                return self._execute_simple_mode_subprocess(code, language, timeout)
 
         except SecurityError as e:
             # Return all failures for security violations
@@ -166,7 +168,7 @@ class CodeExecutor:
         results = []
 
         # Get container
-        container = self.docker_manager.get_or_create_container(language)
+        container = self.docker_manager.get_or_create_container()
 
         try:
             for i, (input_val, expected_output) in enumerate(zip(inputs, outputs)):
@@ -196,9 +198,8 @@ class CodeExecutor:
                     results.append(0)
 
         finally:
-            # Clean up container if needed
-            if settings.docker["remove_containers"]:
-                self.docker_manager.cleanup_container(container)
+            # Keep container running (don't clean up until rm_docker() is called)
+            pass
 
         return results
     
@@ -212,7 +213,7 @@ class CodeExecutor:
         logger.info("Executing in function comparison mode")
         
         # Get container
-        container = self.docker_manager.get_or_create_container(language)
+        container = self.docker_manager.get_or_create_container()
         
         try:
             # Extract function names
@@ -245,8 +246,8 @@ class CodeExecutor:
             return [0]
         
         finally:
-            if settings.docker["remove_containers"]:
-                self.docker_manager.cleanup_container(container)
+            # Keep container running (don't clean up until rm_docker() is called)
+            pass
     
     def _execute_simple_mode(self, code: str, language: str, 
                            timeout: Optional[int]) -> List[int]:
@@ -257,7 +258,7 @@ class CodeExecutor:
         """
         logger.info("Executing in simple mode")
         
-        container = self.docker_manager.get_or_create_container(language)
+        container = self.docker_manager.get_or_create_container()
         
         try:
             result = self._execute_code_in_container(container, code, language, timeout)
@@ -269,8 +270,8 @@ class CodeExecutor:
             return [0]
         
         finally:
-            if settings.docker["remove_containers"]:
-                self.docker_manager.cleanup_container(container)
+            # Keep container running (don't clean up until rm_docker() is called)
+            pass
     
     def _prepare_test_code(self, code: str, input_val: Any, language: str) -> str:
         """Prepare code with input value for testing."""
@@ -281,20 +282,28 @@ class CodeExecutor:
 
 # Test execution
 import sys
+import inspect
+
 try:
-    # Try to find the main function to test
-    import inspect
-    functions = [name for name, obj in globals().items() 
-                if inspect.isfunction(obj) and not name.startswith('_')]
-    
+    # Find all user-defined functions
+    functions = []
+    for name, obj in globals().items():
+        if (inspect.isfunction(obj) and
+            not name.startswith('_') and
+            name not in ['inspect', 'sys']):
+            functions.append(name)
+
     if functions:
         func_name = functions[0]  # Use first function found
         func = globals()[func_name]
+
+        # Call function with input
         result = func({repr(input_val)})
         print(result)
     else:
-        # If no function found, just execute the code
-        exec(compile('{code}', '<string>', 'exec'))
+        print("Error: No function found to test")
+        sys.exit(1)
+
 except Exception as e:
     print(f"Error: {{e}}")
     sys.exit(1)
@@ -389,6 +398,63 @@ except Exception as e:
             # Fallback to string comparison
             return str(actual).strip() == str(expected).strip()
 
+    def _execute_input_output_mode_subprocess(self, inputs: List[Any], outputs: List[Any],
+                                            code: str, language: str,
+                                            timeout: Optional[int]) -> List[int]:
+        """Execute code with input/output validation using subprocess."""
+        results = []
+
+        print(f"[INFO] Executing {len(inputs)} test cases using host system {language}")
+
+        for i, (input_val, expected_output) in enumerate(zip(inputs, outputs)):
+            try:
+                # Execute code with input using subprocess
+                actual_output = self.subprocess_executor.execute_code(
+                    code, language, input_val, timeout
+                )
+
+                # Compare outputs
+                if self._compare_outputs(actual_output, expected_output):
+                    results.append(1)
+                    logger.debug(f"Test case {i+1} passed")
+                else:
+                    results.append(0)
+                    logger.debug(f"Test case {i+1} failed - Expected: {expected_output}, Got: {actual_output}")
+
+            except Exception as e:
+                results.append(0)
+                logger.error(f"Test case {i+1} failed with error: {e}")
+
+        return results
+
+    def _execute_function_comparison_mode_subprocess(self, answer_fn: str, code: str,
+                                                   language: str, timeout: Optional[int]) -> List[int]:
+        """Execute function comparison mode using subprocess."""
+        try:
+            print(f"[INFO] Executing function comparison using host system {language}")
+
+            # For now, just execute the code and return success if no error
+            self.subprocess_executor.execute_code(code, language, None, timeout)
+            return [1]
+
+        except Exception as e:
+            logger.error(f"Function comparison failed: {e}")
+            return [0]
+
+    def _execute_simple_mode_subprocess(self, code: str, language: str,
+                                      timeout: Optional[int]) -> List[int]:
+        """Execute simple mode using subprocess."""
+        try:
+            print(f"[INFO] Executing code using host system {language}")
+
+            # Execute code without input
+            self.subprocess_executor.execute_code(code, language, None, timeout)
+            return [1]
+
+        except Exception as e:
+            logger.error(f"Simple execution failed: {e}")
+            return [0]
+
 
 # Global executor instance
 _executor: Optional[CodeExecutor] = None
@@ -439,3 +505,18 @@ def execute(inputs: Optional[List[Any]] = None,
     """
     executor = get_executor()
     return executor.execute(inputs, outputs, code, answer_fn, language, timeout, mode)
+
+
+def rm_docker() -> None:
+    """Remove all Docker containers and cleanup resources."""
+    try:
+        docker_manager = get_docker_manager()
+        docker_manager.cleanup_all_containers()
+
+        # Also cleanup subprocess executor
+        cleanup_subprocess_executor()
+
+        logger.info("All Docker containers and subprocess resources cleaned up")
+    except Exception as e:
+        logger.error(f"Failed to cleanup resources: {e}")
+        raise
